@@ -26,9 +26,11 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 """
 
+from enum import Enum
 import logging
 import hashlib
 import collections.abc as abc
+from dataclasses import is_dataclass, fields
 
 # Removing this in 2020-12 broke a shocking amount of stuff, such as
 # https://github.com/OP2/PyOP2/pull/605
@@ -215,12 +217,7 @@ class KeyBuilder:
             Now returns the updated *key_hash*.
         """
 
-        digest = None
-
-        try:
-            digest = key._pytools_persistent_hash_digest  # noqa pylint:disable=protected-access
-        except AttributeError:
-            pass
+        digest = getattr(key, "_pytools_persistent_hash_digest", None)
 
         if digest is None:
             try:
@@ -233,7 +230,8 @@ class KeyBuilder:
                 digest = inner_key_hash.digest()
 
         if digest is None:
-            tname = type(key).__name__
+            tp = type(key)
+            tname = tp.__name__
             method = None
             try:
                 method = getattr(self, "update_for_"+tname)
@@ -245,6 +243,12 @@ class KeyBuilder:
                     if isinstance(key, np.dtype):
                         method = self.update_for_specific_dtype
 
+                elif issubclass(tp, Enum):
+                    method = self.update_for_enum
+
+                elif is_dataclass(tp):
+                    method = self.update_for_dataclass
+
             if method is not None:
                 inner_key_hash = self.new_hash()
                 method(inner_key_hash, key)
@@ -256,7 +260,8 @@ class KeyBuilder:
 
         if not isinstance(key, type):
             try:
-                key._pytools_persistent_hash_digest = digest   # noqa pylint:disable=protected-access
+                # pylint:disable=protected-access
+                object.__setattr__(key, "_pytools_persistent_hash_digest",  digest)
             except AttributeError:
                 pass
             except TypeError:
@@ -287,6 +292,10 @@ class KeyBuilder:
             except OverflowError:
                 sz *= 2
 
+    @classmethod
+    def update_for_enum(cls, key_hash, key):
+        cls.update_for_str(key_hash, str(key))
+
     @staticmethod
     def update_for_bool(key_hash, key):
         key_hash.update(str(key).encode("utf8"))
@@ -331,6 +340,11 @@ class KeyBuilder:
     def update_for_specific_dtype(key_hash, key):
         key_hash.update(key.str.encode("utf8"))
 
+    def update_for_dataclass(self, key_hash, key):
+        for fld in fields(key):
+            self.rec(key_hash, fld.name)
+            self.rec(key_hash, getattr(key, fld.name, None))
+
     # }}}
 
 # }}}
